import io
import os
from typing import Optional, List, Union

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin
from platypush.schemas.google.drive import GoogleDriveFileSchema


class GoogleDrivePlugin(GooglePlugin):
    r"""
    Google Drive plugin.

        1. Create your Google application, if you don't have one already, on
           the `developers console <https://console.developers.google.com>`_.

        2. You may have to explicitly enable your user to use the app if the app
           is created in test mode. Go to "OAuth consent screen" and add your user's
           email address to the list of authorized users.

        3. Select the scopes that you want to enable for your application, depending
           on the integrations that you want to use.
           See https://developers.google.com/identity/protocols/oauth2/scopes
           for a list of the available scopes.

        4. Click on "Credentials", then "Create credentials" -> "OAuth client ID".

        5. Select "Desktop app", enter whichever name you like, and click "Create".

        6. Click on the "Download JSON" icon next to your newly created client ID.
           Save the JSON file under
           ``<WORKDIR>/credentials/google/client_secret.json``.

        7. If you're running the service on a desktop environment, then you
           can just start the application. A browser window will open and
           you'll be asked to authorize the application - you may be prompted
           with a warning because you are running a personal and potentially
           unverified application. After authorizing the application, the
           process will save the credentials under
           ``<WORKDIR>/credentials/google/<list,of,scopes>.json`` and proceed
           with the plugin initialization.

        8. If you're running the service on a headless environment, or you
           prefer to manually generate the credentials file before copying to
           another machine, you can run the following command:

                .. code-block:: bash

                  mkdir -p <WORKDIR>/credentials/google
                  python -m platypush.plugins.google.credentials \
                      'drive,drive.appfolder,drive.photos.readonly' \
                      <WORKDIR>/credentials/google/client_secret.json [--noauth_local_webserver]

           When launched with ``--noauth_local_webserver``, the script will
           start a local webserver and print a URL that should be opened in
           your browser. After authorizing the application, you may be
           prompted with a code that you should copy and paste back to the
           script. Otherwise, if you're running the script on a desktop, a
           browser window will be opened automatically.

    """

    scopes = [
        'https://www.googleapis.com/auth/drive',
        'https://www.googleapis.com/auth/drive.appfolder',
        'https://www.googleapis.com/auth/drive.file',
    ]

    def __init__(self, **kwargs):
        super().__init__(scopes=self.scopes, **kwargs)

    def get_service(self, *_, **__):
        return super().get_service(service='drive', version='v3')

    @action
    def files(
        self,
        filter: Optional[str] = None,
        folder_id: Optional[str] = None,
        limit: Optional[int] = 100,
        drive_id: Optional[str] = None,
        spaces: Optional[Union[str, List[str]]] = None,
        order_by: Optional[Union[str, List[str]]] = None,
    ) -> List[dict]:
        """
        Get the list of files.

        :param filter: Optional filter (default: None). See
            `Google Drive API docs <https://developers.google.com/drive/api/v3/search-files>`_ for
            the supported syntax.

        :param folder_id: Drive folder ID to search (default: get all files).
        :param limit: Maximum number of entries to be retrieves (default: 100).
        :param drive_id: Shared drive ID to search (default: None).
        :param spaces: Drive spaces to search. Supported values:

            - ``drive``
            - ``appDataFolder``
            - ``photos``

        :param order_by: Order the results by a specific attribute or list
            of attributes (default: None). Supported attributes:

                - ``createdTime``
                - ``folder``
                - ``modifiedByMeTime``
                - ``modifiedTime``
                - ``name``
                - ``name_natural``
                - ``quotaBytesUsed``
                - ``recency``
                - ``sharedWithMeTime``
                - ``starred``
                - ``viewedByMeTime``

            Attributes will be sorted in ascending order by default. You can change that by
            by appending "``desc``" separated by a space to the attribute you want in descending
            order - e.g. ``["folder", "createdTime desc", "modifiedTime desc"]``.

        :return: .. schema:: google.drive.GoogleDriveFileSchema(many=True)
        """
        service = self.get_service()
        page_token = None
        files = []

        if isinstance(order_by, list):
            order_by = ','.join(order_by)
        if isinstance(spaces, list):
            spaces = ','.join(spaces)
        if folder_id:
            if not filter:
                filter = ''
            else:
                filter += ' '

            filter += f"'{folder_id}' in parents"

        while True:
            results = (
                service.files()
                .list(
                    q=filter,
                    driveId=drive_id,
                    pageSize=limit,
                    orderBy=order_by,
                    fields="nextPageToken, files(id, name, kind, mimeType)",
                    pageToken=page_token,
                    spaces=spaces,
                )
                .execute()
            )

            page_token = results.get('nextPageToken')
            files.extend(
                GoogleDriveFileSchema().dump(results.get('files', []), many=True)
            )

            if not page_token or (limit and len(files) >= limit):
                break

        return files

    @action
    def get(self, file_id: str):
        """
        Get the information of a file on the Drive by file ID.

        :param file_id: File ID.
        :return: .. schema:: google.drive.GoogleDriveFileSchema
        """
        service = self.get_service()
        file = service.files().get(fileId=file_id).execute()
        return GoogleDriveFileSchema().dump(file)

    @action
    def upload(
        self,
        path: str,
        mime_type: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        parents: Optional[List[str]] = None,
        starred: bool = False,
        target_mime_type: Optional[str] = None,
    ) -> dict:
        """
        Upload a file to Google Drive.

        :param path: Path of the file to upload.
        :param mime_type: MIME type of the source file (e.g. "``image/jpeg``").
        :param name: Name of the target file. Default: same name as the source file.
        :param description: File description.
        :param parents: List of folder IDs that will contain the file (default: drive root).
        :param starred: If True, then the uploaded file will be marked as starred by the user.
        :param target_mime_type: Target MIME type. Useful if you want to e.g. import a CSV file as a Google Sheet
            (use "``application/vnd.google-apps.spreadsheet``), or an ODT file to a Google Doc
            (use "``application/vnd.google-apps.document``). See
            `the official documentation <https://developers.google.com/drive/api/v3/mime-types>`_ for a complete list
            of supported types.
        :return: The uploaded file metadata.

          .. schema:: google.drive.GoogleDriveFileSchema

        """
        from googleapiclient.http import MediaFileUpload

        path = os.path.abspath(os.path.expanduser(path))
        name = name or os.path.basename(path)
        metadata = {
            'name': name,
            'description': description,
            'parents': parents,
            'starred': starred,
        }

        if target_mime_type:
            metadata['mimeType'] = target_mime_type

        media = MediaFileUpload(path, mimetype=mime_type)
        service = self.get_service()
        file = (
            service.files()
            .create(body=metadata, media_body=media, fields='*')
            .execute()
        )

        return dict(GoogleDriveFileSchema().dump(file))

    @action
    def download(self, file_id: str, path: str) -> str:
        """
        Download a Google Drive file locally.

        :param file_id: Path of the file to upload.
        :param path: Path of the file to upload.
        :return: The local file path.
        """
        from googleapiclient.http import MediaIoBaseDownload

        service = self.get_service()
        request = service.files().get_media(fileId=file_id)
        path = os.path.abspath(os.path.expanduser(path))
        if os.path.isdir(path):
            name = service.files().get(fileId=file_id).execute().get('name')
            path = os.path.join(path, name)

        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False

        while not done:
            status, done = downloader.next_chunk()
            self.logger.info('Download progress: %s%%', status.progress())

        with open(path, 'wb') as f:
            f.write(fh.getbuffer().tobytes())

        return path

    @action
    def create(
        self,
        name: str,
        description: Optional[str] = None,
        mime_type: Optional[str] = None,
        parents: Optional[List[str]] = None,
        starred: bool = False,
    ) -> dict:
        """
        Create a file.

        :param name: File name.
        :param description: File description.
        :param mime_type: File MIME type.
        :param parents: List of folder IDs that will contain the file (default: drive root).
        :param starred: If True then the file will be marked as starred.
        :return: The created file metadata.

          .. schema:: google.drive.GoogleDriveFileSchema

        """
        metadata = {
            'name': name,
            'description': description,
            'parents': parents,
            'starred': starred,
        }

        if mime_type:
            metadata['mimeType'] = mime_type

        service = self.get_service()
        file = service.files().create(body=metadata, fields='*').execute()
        return dict(GoogleDriveFileSchema().dump(file))

    @action
    def update(
        self,
        file_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        add_parents: Optional[List[str]] = None,
        remove_parents: Optional[List[str]] = None,
        mime_type: Optional[str] = None,
        starred: Optional[bool] = None,
        trashed: Optional[bool] = None,
    ) -> dict:
        """
        Update the metadata or the content of a file.

        :param file_id: File ID.
        :param name: Set the file name.
        :param description: Set the file description.
        :param add_parents: Add the file to these parent folder IDs.
        :param remove_parents: Remove the file from these parent folder IDs.
        :param mime_type: Set the file MIME type.
        :param starred: Change the starred flag.
        :param trashed: Move/remove from trash.
        :return: The updated file metadata.

          .. schema:: google.drive.GoogleDriveFileSchema

        """
        metadata = {}
        if name:
            metadata['name'] = name
        if description is not None:
            metadata['description'] = description
        if add_parents:
            metadata['add_parents'] = add_parents
        if remove_parents:
            metadata['remove_parents'] = remove_parents
        if mime_type:
            metadata['mimeType'] = mime_type
        if starred is not None:
            metadata['starred'] = starred
        if trashed is not None:
            metadata['trashed'] = trashed

        return dict(
            GoogleDriveFileSchema().dump(
                self.get_service()
                .files()
                .update(fileId=file_id, body=metadata, fields='*')
                .execute()
            )
        )

    @action
    def delete(self, file_id: str):
        """
        Delete a file from Google Drive.

        :param file_id: File ID.
        """
        service = self.get_service()
        service.files().delete(fileId=file_id).execute()

    @action
    def copy(self, file_id: str) -> dict:
        """
        Create a copy of a file.

        :param file_id: File ID.
        :return: The copied file metadata.

          .. schema:: google.drive.GoogleDriveFileSchema

        """
        return dict(
            GoogleDriveFileSchema().dump(
                self.get_service().files().copy(fileId=file_id).execute()
            )
        )

    @action
    def empty_trash(self):
        """
        Empty the Drive bin.
        """
        service = self.get_service()
        service.files().emptyTrash()


# vim:sw=4:ts=4:et:



from contextlib import contextmanager
from threading import RLock
from typing import List, Mapping

from platypush.entities.devices import Device
from platypush.entities.distance import DistanceSensor
from platypush.plugins.sensor import SensorPlugin


# pylint: disable=too-many-ancestors
class SensorDistanceVl53l1xPlugin(SensorPlugin):
    """
    Plugin to interact with an `VL53L1x
    <https://www.st.com/en/imaging-and-photonics-solutions/vl53l1x.html>`_
    laser ranger/distance sensor.
    """

    def __init__(self, i2c_bus=1, i2c_address=0x29, poll_interval=3, **kwargs):
        """
        :param i2c_bus: I2C bus number (default: 1)
        :param i2c_address: I2C address (default: 0x29)
        :param poll_interval: How often the integration should poll for new
            measurements (default: 3 seconds).
        """

        super().__init__(poll_interval=poll_interval, **kwargs)

        self.i2c_bus = i2c_bus
        self.i2c_address = i2c_address
        self._device = None
        self._device_lock = RLock()
        self._last_data = {}

    @contextmanager
    def _get_device(self, ranging=1):
        from VL53L1X import VL53L1X  # type: ignore

        with self._device_lock:
            if self._device:
                yield self._device

            self._device = VL53L1X(i2c_bus=self.i2c_bus, i2c_address=self.i2c_address)

            try:
                self._device.open()
                self._device.start_ranging(ranging)
                yield self._device
            finally:
                try:
                    self._device.stop_ranging()
                except Exception:
                    pass

                self._device.close()
                self._device = None

    def get_measurement(self, *_, short=True, medium=True, long=True, **__):
        """
        :param short: Enable short range measurement (default: True)
        :param medium: Enable medium range measurement (default: True)
        :param long: Enable long range measurement (default: True)

        :returns: dict. Example:

        .. code-block:: python

            output = {
                "short": 83,     # Short range measurement in mm
                "medium": 103,   # Medium range measurement in mm
                "long": 200,     # Long range measurement
            }

        """

        ret = {}
        range_idx = 0
        range_name = None
        range_conf = {
            'short': short,
            'medium': medium,
            'long': long,
        }

        for i, (key, enabled) in enumerate(range_conf.items()):
            if not enabled:
                continue

            range_idx = i + 1
            range_name = key

            with self._get_device(ranging=range_idx) as device:
                try:
                    ret[range_name] = device.get_distance()
                except Exception as e:
                    self.logger.exception(e)
                    self.wait_stop(1)

        return ret

    def transform_entities(self, entities: Mapping[str, int]) -> List[Device]:
        return [
            Device(
                id='vl53l1x',
                name='VL53L1X Distance Sensor',
                children=[
                    DistanceSensor(
                        id=f'vl53l1x:{key}',
                        name=f'{key} distance',
                        value=value,
                        unit='mm',
                    )
                    for key, value in entities.items()
                ],
            )
        ]


# vim:sw=4:ts=4:et:



from marshmallow import pre_load

from platypush.schemas.dataclasses import DataClassSchema


class SystemBaseSchema(DataClassSchema):
    """
    Base schema for system info.
    """

    @pre_load
    def pre_load(self, data, **_) -> dict:
        if hasattr(data, '_asdict'):
            data = data._asdict()
        return data



import base64
import datetime
import hashlib
import json
import os
import random
import time
from typing import Optional, Dict

import rsa

from sqlalchemy import Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import make_transient

from platypush.common.db import Base
from platypush.context import get_plugin
from platypush.exceptions.user import (
    InvalidJWTTokenException,
    InvalidCredentialsException,
)
from platypush.utils import get_or_generate_jwt_rsa_key_pair, utcnow


class UserManager:
    """
    Main class for managing platform users
    """

    def __init__(self):
        db_plugin = get_plugin('db')
        assert db_plugin, 'Database plugin not configured'
        self.db = db_plugin
        self.db.create_all(self.db.get_engine(), Base)

    @staticmethod
    def _mask_password(user):
        make_transient(user)
        user.password = None
        return user

    def _get_session(self, *args, **kwargs):
        return self.db.get_session(self.db.get_engine(), *args, **kwargs)

    def get_user(self, username):
        with self._get_session() as session:
            user = self._get_user(session, username)
            if not user:
                return None

            session.expunge(user)
            return self._mask_password(user)

    def get_user_count(self):
        with self._get_session() as session:
            return session.query(User).count()

    def get_users(self):
        with self._get_session() as session:
            return session.query(User).all()

    def create_user(self, username: str, password: str, **kwargs):
        if not username:
            raise ValueError('Invalid or empty username')
        if not password:
            raise ValueError('Please provide a password for the user')

        with self._get_session(locked=True) as session:
            user = self._get_user(session, username)
            if user:
                raise NameError(f'The user {username} already exists')

            password_salt = os.urandom(16)
            hmac_iterations = 100_000
            record = User(
                username=username,
                password=self._encrypt_password(
                    password, password_salt, hmac_iterations
                ),
                password_salt=password_salt.hex(),
                hmac_iterations=hmac_iterations,
                created_at=utcnow(),
                **kwargs,
            )

            session.add(record)
            session.commit()
            user = self._get_user(session, username)

        return self._mask_password(user)

    def update_password(self, username, old_password, new_password):
        with self._get_session(locked=True) as session:
            if not self._authenticate_user(session, username, old_password):
                return False

            user = self._get_user(session, username)
            user.password_salt = user.password_salt or os.urandom(16).hex()
            user.hmac_iterations = user.hmac_iterations or 100_000
            salt = bytes.fromhex(user.password_salt)
            user.password = self._encrypt_password(
                new_password, salt, user.hmac_iterations
            )
            session.commit()
            return True

    def authenticate_user(self, username, password):
        with self._get_session() as session:
            return self._authenticate_user(session, username, password)

    def authenticate_user_session(self, session_token):
        with self._get_session() as session:
            user_session = (
                session.query(UserSession)
                .filter_by(session_token=session_token)
                .first()
            )

            expires_at = (
                user_session.expires_at.replace(tzinfo=datetime.timezone.utc)
                if user_session and user_session.expires_at
                else None
            )

            if not user_session or (expires_at and expires_at < utcnow()):
                return None, None

            user = session.query(User).filter_by(user_id=user_session.user_id).first()
            return self._mask_password(user), user_session

    def delete_user(self, username):
        with self._get_session(locked=True) as session:
            user = self._get_user(session, username)
            if not user:
                raise NameError(f'No such user: {username}')

            user_sessions = (
                session.query(UserSession).filter_by(user_id=user.user_id).all()
            )
            for user_session in user_sessions:
                session.delete(user_session)

            session.delete(user)
            session.commit()
            return True

    def delete_user_session(self, session_token):
        with self._get_session(locked=True) as session:
            user_session = (
                session.query(UserSession)
                .filter_by(session_token=session_token)
                .first()
            )

            if not user_session:
                return False

            session.delete(user_session)
            session.commit()
            return True

    def create_user_session(self, username, password, expires_at=None):
        with self._get_session(locked=True) as session:
            user = self._authenticate_user(session, username, password)
            if not user:
                return None

            if expires_at:
                if isinstance(expires_at, (int, float)):
                    expires_at = datetime.datetime.fromtimestamp(expires_at)
                elif isinstance(expires_at, str):
                    expires_at = datetime.datetime.fromisoformat(expires_at)

            user_session = UserSession(
                user_id=user.user_id,
                session_token=self.generate_session_token(),
                csrf_token=self.generate_session_token(),
                created_at=utcnow(),
                expires_at=expires_at,
            )

            session.add(user_session)
            session.commit()
            return user_session

    @staticmethod
    def _get_user(session, username):
        return session.query(User).filter_by(username=username).first()

    @classmethod
    def _encrypt_password(
        cls, pwd: str, salt: Optional[bytes] = None, iterations: Optional[int] = None
    ) -> str:
        # Legacy password check that uses bcrypt if no salt and iterations are provided
        # See https://git.platypush.tech/platypush/platypush/issues/397
        if not (salt and iterations):
            import bcrypt

            return bcrypt.hashpw(pwd.encode(), bcrypt.gensalt(12)).decode()

        return hashlib.pbkdf2_hmac('sha256', pwd.encode(), salt, iterations).hex()

    @classmethod
    def _check_password(
        cls,
        pwd: str,
        hashed_pwd: str,
        salt: Optional[bytes] = None,
        iterations: Optional[int] = None,
    ) -> bool:
        # Legacy password check that uses bcrypt if no salt and iterations are provided
        # See https://git.platypush.tech/platypush/platypush/issues/397
        if not (salt and iterations):
            import bcrypt

            return bcrypt.checkpw(pwd.encode(), hashed_pwd.encode())

        return (
            hashlib.pbkdf2_hmac(
                'sha256',
                pwd.encode(),
                salt,
                iterations,
            ).hex()
            == hashed_pwd
        )

    @staticmethod
    def _to_bytes(data) -> bytes:
        if isinstance(data, str):
            data = data.encode()
        return data

    @staticmethod
    def generate_session_token():
        rand = bytes(random.randint(0, 255) for _ in range(0, 255))
        return hashlib.sha256(rand).hexdigest()

    def get_user_by_session(self, session_token: str):
        """
        Get a user associated to a session token.

        :param session_token: Session token.
        """
        with self._get_session() as session:
            return (
                session.query(User)
                .join(UserSession)
                .filter_by(session_token=session_token)
                .first()
            )

    def generate_jwt_token(
        self,
        username: str,
        password: str,
        expires_at: Optional[datetime.datetime] = None,
    ) -> str:
        """
        Create a user JWT token for API usage.

        :param username: User name.
        :param password: Password.
        :param expires_at: Expiration datetime of the token.
        :return: The generated JWT token as a string.
        :raises: :class:`platypush.exceptions.user.InvalidCredentialsException` in case of invalid credentials.
        """
        user = self.authenticate_user(username, password)
        if not user:
            raise InvalidCredentialsException()

        pub_key, _ = get_or_generate_jwt_rsa_key_pair()
        payload = json.dumps(
            {
                'username': username,
                'password': password,
                'created_at': datetime.datetime.now().timestamp(),
                'expires_at': expires_at.timestamp() if expires_at else None,
            },
            sort_keys=True,
            indent=None,
        )

        return base64.b64encode(rsa.encrypt(payload.encode('ascii'), pub_key)).decode()

    def validate_jwt_token(self, token: str) -> Dict[str, str]:
        """
        Validate a JWT token.

        :param token: Token to validate.
        :return: On success, it returns the JWT payload with the following structure:

            .. code-block:: json

                {
                    "username": "user ID/name",
                    "created_at": "token creation timestamp",
                    "expires_at": "token expiration timestamp"
                }

        :raises: :class:`platypush.exceptions.user.InvalidJWTTokenException` in case of invalid token.
        """
        _, priv_key = get_or_generate_jwt_rsa_key_pair()

        try:
            payload = json.loads(
                rsa.decrypt(base64.b64decode(token.encode('ascii')), priv_key).decode(
                    'ascii'
                )
            )
        except (TypeError, ValueError) as e:
            raise InvalidJWTTokenException(f'Could not decode JWT token: {e}') from e

        expires_at = payload.get('expires_at')
        if expires_at and time.time() > expires_at:
            raise InvalidJWTTokenException('Expired JWT token')

        user = self.authenticate_user(
            payload.get('username', ''), payload.get('password', '')
        )

        if not user:
            raise InvalidCredentialsException()

        return payload

    def _authenticate_user(self, session, username, password):
        """
        :return: :class:`platypush.user.User` instance if the user exists and the password is valid, ``None`` otherwise.
        """
        user = self._get_user(session, username)
        if not user:
            return None

        if not self._check_password(
            password,
            user.password,
            bytes.fromhex(user.password_salt) if user.password_salt else None,
            user.hmac_iterations,
        ):
            return None

        return user


class User(Base):
    """Models the User table"""

    __tablename__ = 'user'
    __table_args__ = {'sqlite_autoincrement': True}

    user_id = Column(Integer, primary_key=True)
    username = Column(String, unique=True, nullable=False)
    password = Column(String)
    password_salt = Column(String)
    hmac_iterations = Column(Integer)
    created_at = Column(DateTime)


class UserSession(Base):
    """Models the UserSession table"""

    __tablename__ = 'user_session'
    __table_args__ = {'sqlite_autoincrement': True}

    session_id = Column(Integer, primary_key=True)
    session_token = Column(String, unique=True, nullable=False)
    csrf_token = Column(String, unique=True)
    user_id = Column(Integer, ForeignKey('user.user_id'), nullable=False)
    created_at = Column(DateTime)
    expires_at = Column(DateTime)


# vim:sw=4:ts=4:et:



import threading

from typing import IO, Optional, List, Tuple, Union

from platypush.plugins import action
from platypush.plugins.camera import CameraPlugin, Camera
from platypush.utils import wait_for_either

from .model import PiCameraInfo, PiCamera


class CameraPiLegacyPlugin(CameraPlugin):
    """
    Plugin to interact with a `Pi Camera
    <https://www.raspberrypi.com/documentation/accessories/camera.html>`_.

    .. warning::
        This plugin is **DEPRECATED**, as it relies on the old ``picamera``
        module.


    The ``picamera`` module used in this plugin is deprecated and no longer
    maintained. The `picamera2 <https://github.com/raspberrypi/picamera2>`_
    module is advised instead, which is used by
    :class:`platypush.plugins.camera.pi.CameraPiPlugin`.

    You may want to use this plugin if you are running an old OS that does not
    support the new ``picamera2`` module. Even in that case, you may probably
    consider using :class:`platypush.plugins.camera.ffmpeg.FfmpegCameraPlugin`
    or :class:`platypush.plugins.camera.gstreamer.GStreamerCameraPlugin`, as
    ``picamera`` is not maintained anymore and may not work properly.
    """

    _camera_class = PiCamera
    _camera_info_class = PiCameraInfo
    _supported_encoders = ('h264', 'mjpeg')

    def __init__(
        self,
        device: int = 0,
        fps: float = 30.0,
        warmup_seconds: float = 2.0,
        sharpness: int = 0,
        contrast: int = 0,
        brightness: int = 50,
        video_stabilization: bool = False,
        iso: int = 0,
        exposure_compensation: int = 0,
        exposure_mode: str = 'auto',
        meter_mode: str = 'average',
        awb_mode: str = 'auto',
        image_effect: str = 'none',
        led_pin: Optional[int] = None,
        color_effects: Optional[Union[str, List[str]]] = None,
        zoom: Tuple[float, float, float, float] = (0.0, 0.0, 1.0, 1.0),
        stream_format: str = 'h264',
        **camera,
    ):
        """
        :param device: Camera device number (default: 0). Only supported on
            devices with multiple camera slots.
        :param fps: Frames per second (default: 30.0).
        :param warmup_seconds: Seconds to wait for the camera to warm up
            before taking a photo (default: 2.0).
        :param sharpness: Sharpness level, as an integer between -100 and 100.
        :param contrast: Contrast level, as an integer between -100 and 100.
        :param brightness: Brightness level, as an integer between 0 and 100.
        :param video_stabilization: Enable video stabilization (default: False).
        :param iso: ISO level (default: 0).
        :param exposure_compensation: Exposure compensation level, as an
            integer between -25 and 25.
        :param exposure_mode: Exposure mode. Allowed values:

            - ``off``
            - ``auto`` (default)
            - ``night``
            - ``nightpreview``
            - ``backlight``
            - ``spotlight``
            - ``sports``
            - ``snow``
            - ``beach``
            - ``verylong``
            - ``fixedfps``
            - ``antishake``
            - ``fireworks``

        :param meter_mode: Metering mode used for the exposure. Allowed values:

            - ``average`` (default)
            - ``spot``
            - ``backlit``
            - ``matrix``

        :param awb_mode: Auto white balance mode. Allowed values:

            - ``off``
            - ``auto`` (default)
            - ``sunlight``
            - ``cloudy``
            - ``shade``
            - ``tungsten``
            - ``fluorescent``
            - ``incandescent``
            - ``flash``
            - ``horizon``

        :param image_effect: Image effect applied to the camera. Allowed values:

            - ``none`` (default)
            - ``negative``
            - ``solarize``
            - ``sketch``
            - ``denoise``
            - ``emboss``
            - ``oilpaint``
            - ``hatch``
            - ``gpen``
            - ``pastel``
            - ``watercolor``
            - ``film``
            - ``blur``
            - ``saturation``
            - ``colorswap``
            - ``washedout``
            - ``posterise``
            - ``colorpoint``
            - ``colorbalance``
            - ``cartoon``
            - ``deinterlace1``
            - ``deinterlace2``

        :param led_pin: LED PIN number, if the camera LED is wired to a GPIO
            PIN and you want to control it.
        :param zoom: Camera zoom, in the format ``(x, y, width, height)``
            (default: ``(0.0, 0.0, 1.0, 1.0)``).
        :param stream_format: Default format for the output when streamed to a
            network device. Available:

                - ``h264`` (default)
                - ``mjpeg``

        :param camera: Options for the base camera plugin (see
            :class:`platypush.plugins.camera.CameraPlugin`).
        """
        super().__init__(
            device=device,
            fps=fps,
            warmup_seconds=warmup_seconds,
            stream_format=stream_format,
            **camera,
        )

        self.camera_info.sharpness = sharpness  # type: ignore
        self.camera_info.contrast = contrast  # type: ignore
        self.camera_info.brightness = brightness  # type: ignore
        self.camera_info.video_stabilization = video_stabilization  # type: ignore
        self.camera_info.iso = iso  # type: ignore
        self.camera_info.exposure_compensation = exposure_compensation  # type: ignore
        self.camera_info.meter_mode = meter_mode  # type: ignore
        self.camera_info.exposure_mode = exposure_mode  # type: ignore
        self.camera_info.awb_mode = awb_mode  # type: ignore
        self.camera_info.image_effect = image_effect  # type: ignore
        self.camera_info.color_effects = color_effects  # type: ignore
        self.camera_info.zoom = zoom  # type: ignore
        self.camera_info.led_pin = led_pin  # type: ignore

    def prepare_device(self, device: Camera, **_):
        import picamera  # type: ignore

        assert isinstance(device, PiCamera), f'Invalid camera type: {type(device)}'
        camera = picamera.PiCamera(
            camera_num=device.info.device,
            resolution=device.info.resolution,
            framerate=device.info.fps,
            led_pin=device.info.led_pin,
        )

        camera.hflip = device.info.horizontal_flip
        camera.vflip = device.info.vertical_flip
        camera.sharpness = device.info.sharpness
        camera.contrast = device.info.contrast
        camera.brightness = device.info.brightness
        camera.video_stabilization = device.info.video_stabilization
        camera.iso = device.info.iso
        camera.exposure_compensation = device.info.exposure_compensation
        camera.exposure_mode = device.info.exposure_mode
        camera.meter_mode = device.info.meter_mode
        camera.awb_mode = device.info.awb_mode
        camera.image_effect = device.info.image_effect
        camera.color_effects = device.info.color_effects
        camera.rotation = device.info.rotate or 0
        camera.zoom = device.info.zoom

        return camera

    def release_device(self, device: Camera):
        import picamera  # type: ignore

        assert isinstance(device, PiCamera), f'Invalid camera type: {type(device)}'
        if device.object:
            try:
                device.object.stop_recording()
            except (ConnectionError, picamera.PiCameraNotRecording):
                pass

        if device.object and not device.object.closed:
            try:
                device.object.close()
            except (ConnectionError, picamera.PiCameraClosed):
                pass

    def capture_frame(self, device: Camera, *_, **__):
        import numpy as np
        from PIL import Image

        assert device.info.resolution, 'Invalid resolution'
        assert device.object, 'Camera not opened'
        shape = (
            device.info.resolution[1] + (device.info.resolution[1] % 16),
            device.info.resolution[0] + (device.info.resolution[0] % 32),
            3,
        )

        frame = np.empty(shape, dtype=np.uint8)
        device.object.capture(frame, 'rgb')
        return Image.fromarray(frame)

    def start_preview(self, camera: Camera):
        """
        Start camera preview.
        """
        assert camera.object, 'Camera not opened'
        camera.object.start_preview()

    def stop_preview(self, camera: Camera):
        """
        Stop camera preview.
        """
        if not camera.object:
            return

        try:
            camera.object.stop_preview()
        except Exception as e:
            self.logger.warning(str(e))

    @action
    def capture_preview(
        self,
        device: Optional[Union[str, int]] = None,
        duration: Optional[float] = None,
        n_frames: Optional[int] = None,
        **camera,
    ) -> dict:
        camera = self.open_device(device=device, **camera)
        self.start_preview(camera)

        if n_frames:
            duration = n_frames * (camera.info.fps or 0)
        if duration:
            threading.Timer(duration, lambda: self.stop_preview(camera))

        return self.status()  # type: ignore

    def _streaming_loop(self, camera: Camera, stream_format: str, sock: IO, *_, **__):
        from picamera import PiCamera as PiCamera_  # type: ignore

        stream_format = stream_format.lower()
        assert (
            stream_format in self._supported_encoders
        ), f'Invalid stream format: {stream_format}. Supported formats: {", ".join(self._supported_encoders)}'
        assert isinstance(camera, PiCamera), f'Invalid camera type: {type(camera)}'
        assert camera.object and isinstance(
            camera.object, PiCamera_
        ), f'Invalid camera object type: {type(camera.object)}'

        cam = camera.object
        try:
            cam.start_recording(sock, format=stream_format)
            while not wait_for_either(
                camera.stop_stream_event, self._should_stop, timeout=1
            ):
                cam.wait_recording(1)
        except ConnectionError:
            self.logger.info('Client closed connection')
        finally:
            try:
                cam.stop_recording()
                self.stop_streaming()
            except Exception as e:
                self.logger.warning('Could not stop streaming: %s', e)

    def _prepare_stream_writer(self, *_, **__):
        """
        Overrides the base method to do nothing - the stream writer is handled
        by the picamera library.
        """


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Optional


@dataclass
class MediaResource:
    """
    Models a media resource
    """

    resource: str
    url: str
    title: Optional[str] = None
    description: Optional[str] = None
    filename: Optional[str] = None
    image: Optional[str] = None
    duration: Optional[float] = None
    channel: Optional[str] = None
    channel_url: Optional[str] = None
    type: Optional[str] = None
    resolution: Optional[str] = None



import contextlib
import inspect
import os
import re
import textwrap as tw
from dataclasses import dataclass, field
from typing import Type, Optional, Dict, Set

from platypush.utils import (
    get_backend_class_by_name,
    get_plugin_class_by_name,
    get_plugin_name_by_class,
    get_backend_name_by_class,
    get_decorators,
)
from platypush.utils.manifest import Manifest, ManifestType, Dependencies

from .._parser import DocstringParser
from .._serialize import Serializable
from . import Constructor, Action
from .component import Component
from .constants import doc_base_url


@dataclass
class Integration(Component, DocstringParser, Serializable):
    """
    Represents the metadata of an integration (plugin or backend).
    """

    _class_type_re = re.compile(r"^<class '(?P<name>[\w_]+)'>$")

    name: str
    type: Type
    doc: Optional[str] = None
    constructor: Optional[Constructor] = None
    actions: Dict[str, Action] = field(default_factory=dict)
    _manifest: Optional[Manifest] = None
    _skip_manifest: bool = False

    def __post_init__(self):
        """
        Initialize the manifest object.
        """
        if not self._skip_manifest:
            self._init_manifest()

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "type": f"{self.type.__module__}.{self.type.__qualname__}",
            "doc": self.doc,
            "doc_url": self.doc_url,
            "args": {
                **(
                    {name: arg.to_dict() for name, arg in self.constructor.args.items()}
                    if self.constructor
                    else {}
                ),
            },
            "actions": {
                k: {
                    "doc_url": f"{self.doc_url}#{self.cls.__module__}.{self.cls.__qualname__}.{k}",
                    **v.to_dict(),
                }
                for k, v in self.actions.items()
                if self.cls
            },
            "events": {
                f"{e.__module__}.{e.__qualname__}": {
                    "doc": inspect.getdoc(e),
                    "doc_url": f"{doc_base_url}/events/"
                    + ".".join(e.__module__.split(".")[3:])
                    + f".html#{e.__module__}.{e.__qualname__}",
                }
                for e in self.events
            },
            "deps": self.deps.to_dict(),
            "config_snippet": self.config_snippet,
        }

    @classmethod
    def _merge_actions(cls, actions: Dict[str, Action], new_actions: Dict[str, Action]):
        """
        Utility function to merge a new mapping of actions into an existing one.
        """
        for action_name, action in new_actions.items():
            # Set the action if it doesn't exist
            if action_name not in actions:
                actions[action_name] = action

            # Set the action documentation if it's not set
            if action.doc and not actions[action_name].doc:
                actions[action_name].doc = action.doc

            # Merge the parameters
            cls._merge_params(actions[action_name].args, action.args)

    @classmethod
    def _merge_events(cls, events: Set[Type], new_events: Set[Type]):
        """
        Utility function to merge a new mapping of actions into an existing one.
        """
        events.update(new_events)

    @classmethod
    def by_name(cls, name: str) -> "Integration":
        """
        :param name: Integration name.
        :return: A parsed Integration class given its type.
        """
        type = (
            get_backend_class_by_name(".".join(name.split(".")[1:]))
            if name.startswith("backend.")
            else get_plugin_class_by_name(name)
        )
        return cls.by_type(type)

    @classmethod
    def by_type(cls, type: Type, _skip_manifest: bool = False) -> "Integration":
        """
        :param type: Integration type (plugin or backend).
        :param _skip_manifest: Whether we should skip parsing the manifest file for this integration
            (you SHOULDN'T use this flag outside of this class!).
        :return: A parsed Integration class given its type.
        """
        from platypush.backend import Backend
        from platypush.plugins import Plugin

        assert issubclass(
            type, (Plugin, Backend)
        ), f"Expected a Plugin or Backend class, got {type}"

        name = (
            get_plugin_name_by_class(type)
            if issubclass(type, Plugin)
            else "backend." + get_backend_name_by_class(type)
        )

        assert name
        obj = cls(
            name=name,
            type=type,
            doc=cls._expand_rst_extensions(inspect.getdoc(type) or '', type) or None,
            constructor=Constructor.parse(type),
            actions={
                name: Action.parse(getattr(type, name))
                for name in get_decorators(type, climb_class_hierarchy=True).get(
                    "action", []
                )
            },
            _skip_manifest=_skip_manifest,
        )

        for p_type in inspect.getmro(type)[1:]:
            with contextlib.suppress(AssertionError):
                p_obj = cls.by_type(p_type, _skip_manifest=True)
                # Merge constructor parameters
                if obj.constructor and p_obj.constructor:
                    cls._merge_params(obj.constructor.args, p_obj.constructor.args)

                # Merge actions
                cls._merge_actions(obj.actions, p_obj.actions)
                # Merge events
                try:
                    cls._merge_events(obj.events, p_obj.events)
                except FileNotFoundError:
                    pass

        return obj

    @property
    def cls(self) -> Optional[Type]:
        """
        :return: The class of an integration.
        """
        manifest_type = self.manifest.package.split(".")[1]
        if manifest_type == "backend":
            getter = get_backend_class_by_name
        elif manifest_type == "plugins":
            getter = get_plugin_class_by_name
        else:
            return None

        return getter(".".join(self.manifest.package.split(".")[2:]))

    @property
    def base_type(self) -> Type:
        """
        :return: The base type of this integration, either :class:`platypush.backend.Backend` or
            :class:`platypush.plugins.Plugin`.
        """
        from platypush.backend import Backend
        from platypush.plugins import Plugin

        assert self.cls, f'No class found for integration {self.name}'
        if issubclass(self.cls, Plugin):
            return Plugin
        if issubclass(self.cls, Backend):
            return Backend

        raise AssertionError(f"Unknown base type for {self.cls}")

    @classmethod
    def from_manifest(cls, manifest_file: str) -> "Integration":
        """
        Create an `IntegrationMetadata` object from a manifest file.

        :param manifest_file: Path of the manifest file.
        :return: A parsed Integration class given its manifest file.
        """
        manifest = Manifest.from_file(manifest_file)
        name = ".".join(
            [
                "backend" if manifest.manifest_type == ManifestType.BACKEND else "",
                *manifest.package.split(".")[2:],
            ]
        ).strip(".")

        return cls.by_name(name)

    def _init_manifest(self) -> Manifest:
        """
        Initialize the manifest object.
        """
        if not self._manifest:
            self._manifest = Manifest.from_file(self.manifest_file)
        return self._manifest

    @classmethod
    def _type_str(cls, param_type) -> str:
        """
        Utility method to pretty-print the type string of a parameter.
        """
        type_str = str(param_type).replace("typing.", "")
        if m := cls._class_type_re.match(type_str):
            return m.group("name")

        return type_str

    @property
    def manifest(self) -> Manifest:
        """
        :return: The parsed Manifest object.
        """
        return self._init_manifest()

    @property
    def manifest_file(self) -> str:
        """
        :return: Path of the manifest file for the integration.
        """
        return os.path.join(
            os.path.dirname(inspect.getfile(self.type)), "manifest.json"
        )

    @property
    def description(self) -> Optional[str]:
        """
        :return: The description of the integration.
        """
        return self.manifest.description

    @property
    def events(self) -> Set[Type]:
        """
        :return: Events triggered by the integration.
        """
        return set(self.manifest.events)

    @property
    def deps(self) -> Dependencies:
        """
        :return: Dependencies of the integration.
        """
        return self.manifest.install

    @classmethod
    def _indent_yaml_comment(cls, s: str) -> str:
        return tw.indent(
            "\n".join(
                [
                    line if line.startswith("#") else f"# {line}"
                    for line in s.split("\n")
                ]
            ),
            "  ",
        )

    @property
    def config_snippet(self) -> str:
        """
        :return: A YAML snippet with the configuration parameters of the integration.
        """
        return tw.dedent(
            self.name
            + ":\n"
            + (
                "\n".join(
                    f'  # [{"Required" if param.required else "Optional"}]\n'
                    + (f"{self._indent_yaml_comment(param.doc)}" if param.doc else "")
                    + "\n  "
                    + ("# " if not param.required else "")
                    + f"{name}: "
                    + (str(param.default) if param.default is not None else "")
                    + (
                        self._indent_yaml_comment(f"type={self._type_str(param.type)}")
                        if param.type
                        else ""
                    )
                    + "\n"
                    for name, param in self.constructor.args.items()
                )
                if self.constructor and self.constructor.args
                else "    # No configuration required\n"
            )
        )

    @property
    def doc_url(self) -> str:
        """
        :return: URL of the documentation for the integration.
        """
        from platypush.backend import Backend
        from platypush.plugins import Plugin

        if issubclass(self.type, Plugin):
            section = 'plugins'
        elif issubclass(self.type, Backend):
            section = 'backend'
        else:
            raise AssertionError(f'Unknown integration type {self.type}')

        return f"{doc_base_url}/{section}/{self.name}.html"



from abc import ABC, abstractmethod


class Serializable(ABC):
    """
    Base class for reflection entities that can be serialized to JSON/YAML.
    """

    @abstractmethod
    def to_dict(self) -> dict:
        """
        Serialize the entity to a string.
        """
        raise NotImplementedError()



from typing import Union
from platypush.message.event import Event


class KafkaMessageEvent(Event):
    """
    Kafka message event object. Fired when :mod:`platypush.backend.kafka` receives
    a new event.
    """

    def __init__(
        self,
        *args,
        msg: Union[str, list, dict],
        topic: str,
        host: str,
        port: int,
        **kwargs
    ):
        """
        :param msg: Received message. If the message is a JSON string, it will
            be returned as a dict or list. If it's a binary blob, it will be
            returned as a base64-encoded string.
        :param topic: Topic where the message was received.
        :param host: Host where the message was received.
        :param port: Port where the message was received.
        """
        super().__init__(*args, msg=msg, topic=topic, host=host, port=port, **kwargs)


# vim:sw=4:ts=4:et:



from typing import Union

import aioxmpp
import aioxmpp.roster.xso

from platypush.message.event.xmpp import (
    XmppContactAddRequestAcceptedEvent,
    XmppContactAddRequestEvent,
    XmppContactAddRequestRejectedEvent,
)

from .._types import Errors
from ._base import XmppBaseHandler


# pylint: disable=too-many-ancestors
class XmppRosterHandler(XmppBaseHandler):
    """
    Handler for XMPP roster events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.roster: aioxmpp.roster.RosterClient = self._client.summon(
            aioxmpp.roster.RosterClient
        )
        self.roster.on_entry_added.connect(self._on_roster_entry_added)  # type: ignore
        self.roster.on_entry_removed.connect(self._on_roster_entry_removed)  # type: ignore
        self.roster.on_subscribe.connect(self._on_roster_subscribe)  # type: ignore

    def restore_state(self):
        if self._loaded_state.users:
            for user_id in self._loaded_state.users:
                self.add_user(user_id)

    def _on_roster_entry_added(self, item: aioxmpp.roster.Item, *_, **__):
        self.add_user(item.jid)

    def _on_roster_entry_removed(self, item: aioxmpp.roster.Item, *_, **__):
        self.remove_user(item.jid)

    def _on_roster_subscribe(self, stanza: aioxmpp.stanza.StanzaBase, *_, **__):
        def accept():
            self.add_user(jid)
            self.roster.approve(stanza.from_)
            self._state.user_invites.pop(jid, None)
            self._post_user_event(XmppContactAddRequestAcceptedEvent, user_id=jid)

        def reject():
            self._state.user_invites.pop(jid, None)
            self._post_user_event(XmppContactAddRequestRejectedEvent, user_id=jid)

        if not (isinstance(stanza, aioxmpp.Presence) and stanza.to == self._jid):
            return  # Not a contact add request

        jid = str(stanza.from_)
        invite = self._state.user_invites[jid]
        self._post_user_event(XmppContactAddRequestEvent, user_id=stanza.from_)  # type: ignore
        invite.on_accepted = accept
        invite.on_rejected = reject

        if self._config.auto_accept_invites:
            invite.accept()

    def accept_invite(self, user_id: str):
        invite = self._state.user_invites.get(user_id)
        assert invite, Errors.NO_INVITE
        invite.accept()

    def reject_invite(self, user_id: str):
        invite = self._state.user_invites.get(user_id)
        assert invite, Errors.NO_INVITE
        invite.reject()

    @staticmethod
    def _get_jid(user_id: Union[str, aioxmpp.JID]) -> aioxmpp.JID:
        return (
            user_id
            if isinstance(user_id, aioxmpp.JID)
            else aioxmpp.JID.fromstr(user_id)
        ).replace(resource=None)

    def add_user(self, user_id: Union[str, aioxmpp.JID]):
        """
        Subscribe and add a user to the roster.
        """

        async def async_wrapper(*_, **__):
            self.roster.subscribe(jid)
            await self.roster.set_entry(jid)

        jid = self._get_jid(user_id)
        self._async_run(async_wrapper, wait_result=False)
        self._state.users.add(str(jid))
        self._state.user_invites.pop(str(jid), None)

        if self._state_serializer:
            self._state_serializer.enqueue(self._state)

    def remove_user(self, user_id: Union[str, aioxmpp.JID]):
        """
        Remove a user from the roster.
        """

        async def async_wrapper(*_, **__):
            self.roster.unsubscribe(jid)
            await self.roster.remove_entry(jid)

        jid = self._get_jid(user_id)
        self._async_run(async_wrapper, wait_result=False)
        self._state.user_invites.pop(str(jid), None)

        if str(jid) in self._state.users:
            self._state.users.remove(str(jid))
        if self._state_serializer:
            self._state_serializer.enqueue(self._state)



from sqlalchemy import Column, Integer, ForeignKey

from platypush.common.db import is_defined

from .sensors import NumericSensor


if not is_defined('link_quality'):

    class LinkQuality(NumericSensor):
        __tablename__ = 'link_quality'

        def __init__(
            self, *args, unit: str = '%', min: float = 0, max: float = 100, **kwargs
        ):
            super().__init__(*args, min=min, max=max, unit=unit, **kwargs)

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from abc import ABC, abstractmethod
from typing import Type

from ._mixins import XmppAsyncMixin, XmppBaseMixin, XmppConfigMixin, XmppEventStateMixin


class XmppBasePlugin(XmppAsyncMixin, XmppConfigMixin, XmppEventStateMixin, ABC):
    """
    Base interface for the XMPP plugin.
    """

    @abstractmethod
    def register_handler(self, hndl_type: Type[XmppBaseMixin]):
        raise NotImplementedError



import logging
import re
import socket
import time

from threading import Thread, Event as ThreadEvent, get_ident
from typing import Optional, Dict

from platypush import __version__
from platypush.bus import Bus
from platypush.common import ExtensionWithManifest
from platypush.config import Config
from platypush.context import get_backend
from platypush.event import EventGenerator
from platypush.message import Message
from platypush.message.event import Event
from platypush.message.event.zeroconf import (
    ZeroconfServiceAddedEvent,
    ZeroconfServiceRemovedEvent,
)
from platypush.message.request import Request
from platypush.message.response import Response
from platypush.utils import (
    clear_timeout,
    get_backend_name_by_class,
    get_redis,
    get_redis_queue_name_by_message,
    get_remaining_timeout,
    set_timeout,
)


class Backend(Thread, EventGenerator, ExtensionWithManifest):
    """
    Parent class for backends.

    A backend is basically a thread that checks for new events on some channel
    (e.g. a network socket, a queue, some new entries on an API endpoint or an
    RSS feed, a voice command through an assistant, a new measure from a sensor
    etc.) and propagates event messages to the main application bus whenever a
    new event happens. You can then build whichever type of custom logic you
    want on such events.
    """

    _default_response_timeout = 5

    # Loop function, can be implemented by derived classes
    loop = None

    def __init__(
        self, bus: Optional[Bus] = None, poll_seconds: Optional[float] = None, **kwargs
    ):
        """
        :param bus: Reference to the bus object to be used in the backend
        :param poll_seconds: If the backend implements a ``loop`` method, this parameter expresses how often the
            loop should run in seconds.
        :param kwargs: Key-value configuration for the backend
        """

        self._thread_name = self.__class__.__name__
        EventGenerator.__init__(self)
        ExtensionWithManifest.__init__(self)
        Thread.__init__(self, name=self._thread_name, daemon=True)

        # If no bus is specified, create an internal queue where
        # the received messages will be pushed
        self.bus = bus or Bus()
        self.poll_seconds = float(poll_seconds) if poll_seconds else None
        self.device_id = Config.get('device_id')
        self.thread_id = None
        self._stop_event = ThreadEvent()
        self._stop_thread: Optional[Thread] = None
        self._kwargs = kwargs
        self.logger = logging.getLogger(
            'platypush:backend:' + get_backend_name_by_class(self.__class__)
        )
        self.zeroconf = None
        self.zeroconf_info = None

        # Internal-only, we set the request context on a backend if that
        # backend is intended to react for a response to a specific request
        self._request_context = kwargs['_req_ctx'] if '_req_ctx' in kwargs else None

        if 'logging' in kwargs:
            self.logger.setLevel(getattr(logging, kwargs['logging'].upper()))

    def on_message(self, msg):
        """
        Callback when a message is received on the backend.
        It parses and posts the message on the main bus.
        It should be called by the derived classes whenever
        a new message should be processed.

        :param msg: Received message.  It can be either a key-value dictionary, a platypush.message.Message object,
            or a string/byte UTF-8 encoded string
        """

        msg = Message.build(msg)
        if msg is None:
            return

        if not getattr(msg, 'target', None) or msg.target != self.device_id:
            return  # Not for me

        self.logger.debug(
            'Message received on the %s backend: %s', self.__class__.__name__, msg
        )

        if self._is_expected_response(msg):
            # Expected response, trigger the response handler
            clear_timeout()
            if self._request_context:
                # pylint: disable=unsubscriptable-object
                self._request_context['on_response'](msg)

            self.stop()
            return

        msg.backend = self  # Augment message to be able to process responses
        self.bus.post(msg)

    def _is_expected_response(self, msg):
        """
        Internal only - returns true if we are expecting for a response and msg
        is that response.
        """

        # pylint: disable=unsubscriptable-object
        return (
            self._request_context
            and isinstance(msg, Response)
            and msg.id == self._request_context['request'].id
        )

    def _get_backend_config(self):
        config_name = (
            'backend.' + self.__class__.__name__.split('Backend', maxsplit=1)[0].lower()
        )
        return Config.get(config_name) or {}

    def _setup_response_handler(self, request, on_response, response_timeout):
        def _timeout_hndl():
            raise RuntimeError(
                f'Timed out while waiting for a response from {request.target}'
            )

        req_ctx = {
            'request': request,
            'on_response': on_response,
            'response_timeout': response_timeout,
        }

        resp_backend = self.__class__(
            bus=self.bus, _req_ctx=req_ctx, **self._get_backend_config(), **self._kwargs
        )

        # Set the response timeout
        if response_timeout:
            set_timeout(seconds=response_timeout, on_timeout=_timeout_hndl)

        resp_backend.start()

    def send_event(self, event, **kwargs):
        """
        Send an event message on the backend.

        :param event: Event to send. It can be a dict, a string/bytes UTF-8 JSON, or a platypush.message.event.Event
            object.
        """

        event = Event.build(event)
        assert isinstance(event, Event)

        event.origin = self.device_id
        if not hasattr(event, 'target'):
            event.target = self.device_id

        self.send_message(event, **kwargs)

    def send_request(
        self,
        request,
        on_response=None,
        response_timeout=_default_response_timeout,
        **kwargs,
    ):
        """
        Send a request message on the backend.

        :param request: The request, either a dict, a string/bytes UTF-8 JSON, or a platypush.message.request.Request
            object.

        :param on_response: Optional callback that will be called when a response is received. If set, this method will
            synchronously wait for a response before exiting.
        :type on_response: function

        :param response_timeout: If on_response is set, the backend will raise an exception if the response isn't
            received within this number of seconds (default: None)
        :type response_timeout: float
        """

        request = Request.build(request)
        assert isinstance(request, Request)

        request.origin = self.device_id

        if on_response and response_timeout != 0:
            self._setup_response_handler(request, on_response, response_timeout)

        self.send_message(request, **kwargs)

    def send_response(self, response, request, **kwargs):
        """
        Send a response message on the backend.

        :param response: The response, either a dict, a string/bytes UTF-8 JSON, or a
            :class:`platypush.message.response.Response` object.
        :param request: Associated request, used to set the response parameters that will link them
        """

        assert isinstance(response, Response)
        assert isinstance(request, Request)

        self.send_message(response, **kwargs)

    def send_message(self, msg, queue_name=None, **_):
        """
        Sends a platypush.message.Message to a node.
        To be implemented in the derived classes. By default, if the Redis
        backend is configured then it will try to deliver the message to
        other consumers through the configured Redis main queue.

        :param msg: The message to send
        :param queue_name: Send the message on a specific queue (default: the queue_name configured on the Redis
            backend)
        """

        try:
            redis = get_backend('redis')
            if not redis:
                raise KeyError()
        except KeyError:
            self.logger.warning(
                (
                    "Backend %s does not implement send_message "
                    "and the fallback Redis backend isn't configured"
                ),
                self.__class__.__name__,
            )
            return

        redis.send_message(msg, queue_name=queue_name)

    def run(self):
        """Starts the backend thread. To be implemented in the derived classes if the loop method isn't defined."""
        self.thread_id = get_ident()
        if not callable(self.loop):
            return

        while not self.should_stop():
            try:
                with self:
                    has_error = False

                    while not self.should_stop() and not has_error:
                        try:
                            # pylint: disable=not-callable
                            self.loop()
                        except Exception as e:
                            has_error = True
                            self.logger.error(str(e))
                            self.logger.exception(e)
                        finally:
                            if self.poll_seconds:
                                time.sleep(self.poll_seconds)
                            elif has_error:
                                time.sleep(5)
            except Exception as e:
                self.logger.error(
                    '%s initialization error: %s', self.__class__.__name__, e
                )
                self.logger.exception(e)
                time.sleep(self.poll_seconds or 5)

    def __enter__(self):
        """Invoked when the backend is initialized, if the main logic is within a ``loop()`` function"""
        self.logger.info('Initialized backend %s', self.__class__.__name__)

    def __exit__(self, *_, **__):
        """Invoked when the backend is terminated, if the main logic is within a ``loop()`` function"""
        self.on_stop()
        self.logger.info('Terminated backend %s', self.__class__.__name__)

    def on_stop(self):
        """Callback invoked when the process stops"""

    def stop(self):
        """Stops the backend thread by sending a STOP event on its bus"""

        def _async_stop():
            self._stop_event.set()
            self.unregister_service()
            self.on_stop()
            self._stop_thread = None

        if not (self._stop_thread and self._stop_thread.is_alive()):
            self._stop_thread = Thread(target=_async_stop)
            self._stop_thread.start()

    def should_stop(self):
        """
        :return: True if the backend thread should be stopped, False otherwise.
        """
        return self._stop_event.is_set()

    def wait_stop(self, timeout=None) -> bool:
        """
        Waits for the backend thread to stop.

        :param timeout: The maximum time to wait for the backend thread to stop (default: None)
        :return: True if the backend thread has stopped, False otherwise.
        """
        start = time.time()

        if self._stop_thread:
            try:
                self._stop_thread.join(
                    get_remaining_timeout(timeout=timeout, start=start)
                )
            except AttributeError:
                pass

        return self._stop_event.wait(
            get_remaining_timeout(timeout=timeout, start=start)
        )

    def get_message_response(self, msg):
        queue = get_redis_queue_name_by_message(msg)
        if not queue:
            self.logger.warning('No response queue configured for the message')
            return None

        try:
            redis = get_redis()
            response = redis.blpop(queue, timeout=60)
            if response and len(response) > 1:
                response = Message.build(response[1])
            else:
                response = None

            return response
        except Exception as e:
            self.logger.error('Error while processing response to %s: %s', msg, e)

        return None

    @staticmethod
    def _get_ip() -> str:
        """
        Get the IP address of the machine.
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        addr = s.getsockname()[0]
        s.close()
        return addr

    def register_service(
        self,
        port: Optional[int] = None,
        name: Optional[str] = None,
        srv_type: Optional[str] = None,
        srv_name: Optional[str] = None,
        udp: bool = False,
        properties: Optional[Dict] = None,
    ):
        """
        Initialize the Zeroconf service configuration for this backend.

        :param port: Service listen port (default: the backend ``port`` attribute if available, or ``None``).
        :param name: Service short name (default: backend name).
        :param srv_type: Service type (default: ``_platypush-{name}._{proto}.local.``).
        :param srv_name: Full service name (default: ``{hostname or device_id}.{type}``).
        :param udp: Set to True if this is a UDP service.
        :param properties: Extra properties to be passed on the service. Default:

            .. code-block:: json

                {
                    "name": "Platypush",
                    "vendor": "Platypush",
                    "version": "{platypush_version}"
                }

        """
        try:
            from zeroconf import ServiceInfo, Zeroconf
            from platypush.plugins.zeroconf import ZeroconfListener
        except ImportError:
            self.logger.warning(
                'zeroconf package not available, service discovery will be disabled.'
            )
            return

        if self.zeroconf:
            self.logger.info(
                'Zeroconf service already registered for %s, removing the previous instance',
                self.__class__.__name__,
            )
            self.unregister_service()

        self.zeroconf = Zeroconf()
        srv_desc = {
            'name': 'Platypush',
            'vendor': 'Platypush',
            'version': __version__,
            **(properties or {}),
        }

        name = name or re.sub(r'Backend$', '', self.__class__.__name__).lower()
        srv_type = srv_type or f'_platypush-{name}._{"udp" if udp else "tcp"}.local.'
        srv_name = srv_name or f'{self.device_id}.{srv_type}'
        srv_port = port if port else getattr(self, 'port', None)
        self.zeroconf_info = ServiceInfo(
            srv_type,
            srv_name,
            addresses=[socket.inet_aton(self._get_ip())],
            port=srv_port,
            weight=0,
            priority=0,
            properties=srv_desc,
        )

        if not self.zeroconf_info:
            self.logger.warning('Could not register Zeroconf service')
            return

        self.zeroconf.register_service(self.zeroconf_info)
        self.bus.post(
            ZeroconfServiceAddedEvent(
                service_type=srv_type,
                service_name=srv_name,
                service_info=ZeroconfListener.parse_service_info(self.zeroconf_info),
            )
        )

    def unregister_service(self):
        """
        Unregister the Zeroconf service configuration if available.
        """
        from redis import exceptions

        if self.zeroconf and self.zeroconf_info:
            try:
                self.zeroconf.unregister_service(self.zeroconf_info)
            except Exception as e:
                self.logger.warning(
                    'Could not register Zeroconf service %s: %s: %s',
                    self.zeroconf_info.name,
                    type(e).__name__,
                    e,
                )

            if self.zeroconf:
                try:
                    self.zeroconf.close()
                except TimeoutError:
                    pass

            try:
                if self.zeroconf_info:
                    self.bus.post(
                        ZeroconfServiceRemovedEvent(
                            service_type=self.zeroconf_info.type,
                            service_name=self.zeroconf_info.name,
                        )
                    )
                else:
                    self.bus.post(
                        ZeroconfServiceRemovedEvent(
                            service_type=None, service_name=None
                        )
                    )
            except exceptions.ConnectionError:
                pass

            self.zeroconf_info = None
            self.zeroconf = None


# vim:sw=4:ts=4:et:



from enum import IntEnum
import logging
import os
import threading
from typing import Any, Callable, Dict, Iterable, Optional, Union

import paho.mqtt.client as mqtt

from platypush.config import Config

MqttCallback = Callable[["MqttClient", Any, mqtt.MQTTMessage], Any]
DEFAULT_TIMEOUT: int = 30


class MqttClient(mqtt.Client, threading.Thread):
    """
    Wrapper class for an MQTT client executed in a separate thread.
    """

    def __init__(
        self,
        *args,
        host: str,
        port: int,
        client_id: str,
        topics: Iterable[str] = (),
        on_message: Optional[MqttCallback] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        tls_cafile: Optional[str] = None,
        tls_certfile: Optional[str] = None,
        tls_keyfile: Optional[str] = None,
        tls_version: Optional[Union[str, IntEnum]] = None,
        tls_ciphers: Optional[str] = None,
        tls_insecure: bool = False,
        timeout: int = DEFAULT_TIMEOUT,
        **kwargs,
    ):
        self.client_id = client_id or str(Config.get('device_id'))
        kwargs['client_id'] = self.client_id

        # Breaking change in paho.mqtt >= 2.0.0: the callback API version
        # parameter should be passed, see
        # https://github.com/eclipse/paho.mqtt.python/blob/28aa2e6b26a86e4b29126323892fb5f43637d6d6/ChangeLog.txt#L6
        cbApiVersion = getattr(mqtt, 'CallbackAPIVersion', None)
        if cbApiVersion:
            kwargs['callback_api_version'] = cbApiVersion.VERSION1

        mqtt.Client.__init__(self, *args, **kwargs)

        threading.Thread.__init__(self, name=f'MQTTClient:{self.client_id}')

        self.logger = logging.getLogger(self.__class__.__name__)
        self.host = host
        self.port = port
        self.tls_cafile = self._expandpath(tls_cafile)
        self.tls_certfile = self._expandpath(tls_certfile)
        self.tls_keyfile = self._expandpath(tls_keyfile)
        self.tls_version = self._get_tls_version(tls_version)
        self.tls_ciphers = self._expandpath(tls_ciphers)
        self.tls_insecure = tls_insecure
        self.username = username
        self.password = password
        self.topics = set(topics or [])
        self.timeout = timeout
        self.on_connect = self.connect_hndl()
        self.on_disconnect = self.disconnect_hndl()

        if on_message:
            self.on_message = on_message  # type: ignore

        if username and password:
            self.username_pw_set(username, password)

        if tls_cafile:
            self.tls_set(
                ca_certs=self.tls_cafile,
                certfile=self.tls_certfile,
                keyfile=self.tls_keyfile,
                tls_version=self.tls_version,
                ciphers=self.tls_ciphers,
            )

            self.tls_insecure_set(self.tls_insecure)

        self._running = False
        self._stop_scheduled = False

    @staticmethod
    def _expandpath(path: Optional[str] = None) -> Optional[str]:
        """
        Utility method to expand a path string.
        """
        return os.path.abspath(os.path.expanduser(path)) if path else None

    @staticmethod
    def _get_tls_version(version: Optional[Union[str, IntEnum]] = None):
        """
        A utility method that normalizes an SSL version string or enum to a
        standard ``_SSLMethod`` enum.
        """
        import ssl

        if not version:
            return None

        if isinstance(version, type(ssl.PROTOCOL_TLS)):
            return version

        if isinstance(version, str):
            version = version.lower()

        if version == 'tls':
            return ssl.PROTOCOL_TLS
        if version == 'tlsv1':
            return ssl.PROTOCOL_TLSv1
        if version == 'tlsv1.1':
            return ssl.PROTOCOL_TLSv1_1
        if version == 'tlsv1.2':
            return ssl.PROTOCOL_TLSv1_2

        raise AssertionError(f'Unrecognized TLS version: {version}')

    def connect(
        self,
        *args,
        host: Optional[str] = None,
        port: Optional[int] = None,
        keepalive: Optional[int] = None,
        **kwargs,
    ):
        """
        Overrides the default connect method.
        """
        if not self.is_connected():
            self.logger.debug(
                'Connecting to MQTT broker %s:%d, client_id=%s...',
                self.host,
                self.port,
                self.client_id,
            )

            return super().connect(
                host=host or self.host,
                port=port or self.port,
                keepalive=keepalive or self.timeout,
                *args,
                **kwargs,
            )

        return None

    @property
    def configuration(self) -> Dict[str, Any]:
        """
        :return: The configuration of the client.
        """
        return {
            'host': self.host,
            'port': self.port,
            'topics': self.topics,
            'on_message': self.on_message,
            'username': self.username,
            'password': self.password,
            'client_id': self.client_id,
            'tls_cafile': self.tls_cafile,
            'tls_certfile': self.tls_certfile,
            'tls_keyfile': self.tls_keyfile,
            'tls_version': self.tls_version,
            'tls_ciphers': self.tls_ciphers,
            'tls_insecure': self.tls_insecure,
            'timeout': self.timeout,
        }

    def subscribe(self, *topics, **kwargs):
        """
        Client subscription handler.
        """
        if not topics:
            topics = self.topics

        self.topics.update(topics)
        for topic in topics:
            super().subscribe(topic, **kwargs)

    def unsubscribe(self, *topics, **kwargs):
        """
        Client unsubscribe handler.
        """
        if not topics:
            topics = self.topics

        for topic in topics:
            if topic not in self.topics:
                self.logger.info('The topic %s is not subscribed', topic)
                continue

            super().unsubscribe(topic, **kwargs)
            self.topics.remove(topic)

    def connect_hndl(self):
        """
        When the client connects, subscribe to all the registered topics.
        """

        def handler(*_, **__):
            self.logger.debug(
                'Connected to MQTT broker %s:%d, client_id=%s',
                self.host,
                self.port,
                self.client_id,
            )
            self.subscribe()

        return handler

    def disconnect_hndl(self):
        """
        Notifies the client disconnection.
        """

        def handler(*_, **__):
            self.logger.debug(
                'Disconnected from MQTT broker %s:%d, client_id=%s',
                self.host,
                self.port,
                self.client_id,
            )

        return handler

    def run(self):
        """
        Connects to the MQTT server, subscribes to all the registered topics
        and listens for messages.
        """
        super().run()
        self.connect()
        self._running = True
        self.loop_forever()

    def stop(self):
        """
        The stop method schedules the stop and disconnects the client.
        """
        if not self.is_alive():
            return

        try:
            self.loop_stop()
        except Exception as e:
            self.logger.debug('Could not stop client loop: %s: %s', type(e).__name__, e)

        self._stop_scheduled = True
        self.disconnect()
        self._running = False


# vim:sw=4:ts=4:et:



import json
import os
from threading import RLock
from typing import Iterable, Optional

from google.auth import jwt
from google.cloud import pubsub_v1 as pubsub  # pylint: disable=no-name-in-module
from google.api_core.exceptions import AlreadyExists, NotFound

from platypush.config import Config
from platypush.message.event.google.pubsub import GooglePubsubMessageEvent
from platypush.plugins import RunnablePlugin, action


class GooglePubsubPlugin(RunnablePlugin):
    """
    Publishes and subscribes to Google Pub/Sub topics.

    You'll need a Google Cloud active project and a set of credentials to use
    this plugin:

        1. Create a project on the `Google Cloud console
           <https://console.cloud.google.com/projectcreate>`_ if you don't have
           one already.

        2. In the `Google Cloud API console
           <https://console.cloud.google.com/apis/credentials/serviceaccountkey>`_
           create a new service account key. Select "New Service Account", choose
           the role "Pub/Sub Editor" and leave the key type as JSON.

        3. Download the JSON service credentials file. By default Platypush
           will look for the credentials file under
           ``<WORKDIR>/credentials/google/pubsub.json``.

    """

    publisher_audience = 'https://pubsub.googleapis.com/google.pubsub.v1.Publisher'
    subscriber_audience = 'https://pubsub.googleapis.com/google.pubsub.v1.Subscriber'
    default_credentials_file = os.path.join(
        Config.get_workdir(), 'credentials', 'google', 'pubsub.json'
    )

    def __init__(
        self,
        credentials_file: str = default_credentials_file,
        topics: Iterable[str] = (),
        **kwargs,
    ):
        """
        :param credentials_file: Path to the JSON credentials file for Google
            pub/sub (default: ``~/.credentials/platypush/google/pubsub.json``)
        :param topics: List of topics to subscribe. You can either specify the
            full topic name in the format
            ``projects/<project_id>/topics/<topic_name>``, where
            ``<project_id>`` must be the ID of your Google Pub/Sub project, or
            just ``<topic_name>`` - in such case it's implied that you refer to
            the ``topic_name`` under the ``project_id`` of your service
            credentials.
        """
        super().__init__(**kwargs)
        self.credentials_file = credentials_file
        self.project_id = self.get_project_id()
        self.topics = topics
        self._subscriber: Optional[pubsub.SubscriberClient] = None
        self._subscriber_lock = RLock()

    def get_project_id(self):
        with open(self.credentials_file) as f:
            return json.load(f).get('project_id')

    def get_credentials(self, audience: str):
        return jwt.Credentials.from_service_account_file(
            self.credentials_file, audience=audience
        )

    def _norm_topic(self, topic: str):
        if not topic.startswith(f'projects/{self.project_id}/topics/'):
            topic = f'projects/{self.project_id}/topics/{topic}'
        return topic

    @action
    def publish(self, topic: str, msg, **kwargs):
        """
        Publish a message to a topic

        :param topic: Topic/channel where the message will be delivered. You
            can either specify the full topic name in the format
            ``projects/<project_id>/topics/<topic_name>``, where
            ``<project_id>`` must be the ID of your Google Pub/Sub project, or
            just ``<topic_name>`` - in such case it's implied that you refer to
            the ``topic_name`` under the ``project_id`` of your service
            credentials.
        :param msg: Message to be sent. It can be a list, a dict, or a Message object
        :param kwargs: Extra arguments to be passed to .publish()
        """
        credentials = self.get_credentials(self.publisher_audience)
        publisher = pubsub.PublisherClient(credentials=credentials)
        topic = self._norm_topic(topic)

        try:
            publisher.create_topic(name=topic)
        except AlreadyExists:
            self.logger.debug('Topic %s already exists', topic)

        if isinstance(msg, (int, float)):
            msg = str(msg)
        if isinstance(msg, (dict, list)):
            msg = json.dumps(msg)
        if isinstance(msg, str):
            msg = msg.encode()

        publisher.publish(topic, msg, **kwargs)

    @action
    def send_message(self, topic: str, msg, **kwargs):
        """
        Alias for :meth:`.publish`
        """
        self.publish(topic=topic, msg=msg, **kwargs)

    @action
    def subscribe(self, topic: str):
        """
        Subscribe to a topic.

        :param topic: Topic/channel where the message will be delivered. You
            can either specify the full topic name in the format
            ``projects/<project_id>/topics/<topic_name>``, where
            ``<project_id>`` must be the ID of your Google Pub/Sub project, or
            just ``<topic_name>`` - in such case it's implied that you refer to
            the ``topic_name`` under the ``project_id`` of your service
            credentials.
        """
        assert self._subscriber, 'Subscriber not initialized'
        topic = self._norm_topic(topic)
        subscription_name = '/'.join(
            [*topic.split('/')[:2], 'subscriptions', topic.split('/')[-1]]
        )

        try:
            self._subscriber.create_subscription(name=subscription_name, topic=topic)
        except AlreadyExists:
            self.logger.debug('Subscription %s already exists', subscription_name)

        self._subscriber.subscribe(subscription_name, self._message_callback(topic))

    @action
    def unsubscribe(self, topic: str):
        """
        Unsubscribe from a topic.

        :param topic: Topic/channel where the message will be delivered. You
            can either specify the full topic name in the format
            ``projects/<project_id>/topics/<topic_name>``, where
            ``<project_id>`` must be the ID of your Google Pub/Sub project, or
            just ``<topic_name>`` - in such case it's implied that you refer to
            the ``topic_name`` under the ``project_id`` of your service
            credentials.
        """
        assert self._subscriber, 'Subscriber not initialized'
        topic = self._norm_topic(topic)
        subscription_name = '/'.join(
            [*topic.split('/')[:2], 'subscriptions', topic.split('/')[-1]]
        )

        try:
            self._subscriber.delete_subscription(subscription=subscription_name)
        except NotFound:
            self.logger.debug('Subscription %s not found', subscription_name)

    def _message_callback(self, topic):
        def callback(msg):
            data = msg.data.decode()
            try:
                data = json.loads(data)
            except Exception as e:
                self.logger.debug('Not a valid JSON: %s: %s', data, e)

            msg.ack()
            self._bus.post(GooglePubsubMessageEvent(topic=topic, msg=data))

        return callback

    def main(self):
        credentials = self.get_credentials(self.subscriber_audience)
        with self._subscriber_lock:
            self._subscriber = pubsub.SubscriberClient(credentials=credentials)

        for topic in self.topics:
            self.subscribe(topic=topic)

        self.wait_stop()
        with self._subscriber_lock:
            self._close()

    def _close(self):
        with self._subscriber_lock:
            if self._subscriber:
                try:
                    self._subscriber.close()
                except Exception as e:
                    self.logger.debug('Error while closing the subscriber: %s', e)

                self._subscriber = None

    def stop(self):
        self._close()
        super().stop()


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from collections import defaultdict
from threading import RLock
from typing import Any, Dict, Iterable, Optional, Tuple, Union

from platypush.entities.bluetooth import BluetoothDevice

from ._model import MajorDeviceClass


class BaseCache(ABC):
    """
    Base cache class for Bluetooth devices and entities.
    """

    _by_address: Dict[str, Any]
    """ Device cache by address. """
    _by_name: Dict[str, Any]
    """ Device cache by name. """

    def __init__(self):
        self._by_address = {}
        self._by_name = {}
        self._insert_locks = defaultdict(RLock)
        """ Locks for inserting devices into the cache. """

    @property
    @abstractmethod
    def _address_field(self) -> str:
        """
        Name of the field that contains the address of the device.
        """

    @property
    @abstractmethod
    def _name_field(self) -> str:
        """
        Name of the field that contains the name of the device.
        """

    def get(self, device: str) -> Optional[Any]:
        """
        Get a device by address or name.
        """
        dev = self._by_address.get(device)
        if not dev:
            dev = self._by_name.get(device)
        return dev

    def add(self, device: Any) -> Any:
        """
        Cache a device.
        """
        with self._insert_locks[device.address]:
            addr = getattr(device, self._address_field)
            name = getattr(device, self._name_field)
            self._by_address[addr] = device
            if name:
                self._by_name[name] = device

        return device

    def keys(self) -> Iterable[str]:
        """
        :return: All the cached device addresses.
        """
        return list(self._by_address.keys())

    def values(self) -> Iterable[Any]:
        """
        :return: All the cached device entities.
        """
        return list(self._by_address.values())

    def items(self) -> Iterable[Tuple[str, Any]]:
        """
        :return: All the cached items, as ``(address, device)`` tuples.
        """
        return list(self._by_address.items())

    def __contains__(self, device: str) -> bool:
        """
        :return: ``True`` if the entry is cached, ``False`` otherwise.
        """
        return self.get(device) is not None


class EntityCache(BaseCache):
    """
    Cache used to store scanned Bluetooth devices as :class:`BluetoothDevice`.
    """

    _by_address: Dict[str, BluetoothDevice]
    _by_name: Dict[str, BluetoothDevice]

    @property
    def _address_field(self) -> str:
        return 'address'

    @property
    def _name_field(self) -> str:
        return 'name'

    def get(self, device: Union[str, BluetoothDevice]) -> Optional[BluetoothDevice]:
        dev_filter = device.address if isinstance(device, BluetoothDevice) else device
        return super().get(dev_filter)

    def add(self, device: BluetoothDevice) -> BluetoothDevice:
        with self._insert_locks[device.address]:
            existing_device = self.get(device)
            if existing_device:
                self._merge_properties(device, existing_device)
                self._merge_children(device, existing_device)
                device = existing_device

            return super().add(device)

    def values(self) -> Iterable[BluetoothDevice]:
        return super().values()

    def items(self) -> Iterable[Tuple[str, BluetoothDevice]]:
        return super().items()

    def __contains__(self, device: Union[str, BluetoothDevice]) -> bool:
        """
        Override the default ``__contains__`` to support lookup by partial
        :class:`BluetoothDevice` objects.
        """
        return super().__contains__(device)

    def _merge_properties(
        self, device: BluetoothDevice, existing_device: BluetoothDevice
    ):
        """
        Coalesce the properties of the two device representations.
        """
        # Coalesce the major device class
        if existing_device.major_device_class == MajorDeviceClass.UNKNOWN:
            existing_device.major_device_class = device.major_device_class

        # Coalesce the other device and service classes
        for attr in ('major_service_classes', 'minor_device_classes'):
            setattr(
                existing_device,
                attr,
                list(
                    {
                        *getattr(existing_device, attr, []),
                        *getattr(device, attr, []),
                    }
                ),
            )

        # Coalesce mutually exclusive supports_* flags
        for attr in ('supports_ble', 'supports_legacy'):
            if not getattr(existing_device, attr, None):
                setattr(existing_device, attr, getattr(device, attr, None) or False)

        # Merge the connected property
        existing_device.connected = (
            device.connected
            if device.connected is not None
            else existing_device.connected
        )

        # Coalesce other manager-specific properties
        for attr in ('rssi', 'tx_power'):
            if getattr(existing_device, attr, None) is None:
                setattr(existing_device, attr, getattr(device, attr, None))

        # Coalesce the reachable flag
        if device.reachable is not None:
            existing_device.reachable = device.reachable

        # Merge the data and meta dictionaries
        for attr in ('data', 'meta'):
            setattr(
                existing_device,
                attr,
                {
                    **(getattr(existing_device, attr) or {}),
                    **(getattr(device, attr) or {}),
                },
            )

    def _merge_children(
        self, device: BluetoothDevice, existing_device: BluetoothDevice
    ):
        """
        Merge the device's children upon set without overwriting the
        existing ones.
        """
        # Map of the existing children
        existing_children = {
            child.external_id: child for child in existing_device.children
        }

        # Map of the new children
        new_children = {child.id: child for child in device.children}

        # Merge the existing children with the new ones without overwriting them
        existing_children.update(new_children)
        existing_device.children = list(existing_children.values())



import inspect
import json
from abc import ABC, abstractmethod
from typing import Any, Optional, Dict, Collection, Type

from platypush.config import Config
from platypush.entities._base import Entity, EntitySavedCallback
from platypush.utils import get_plugin_name_by_class, get_redis, utcnow

_entity_registry_varname = '_platypush/plugin_entity_registry'


class EntityManager(ABC):
    """
    Base mixin for all the integrations that support entities mapping.

    The classes that implement the entity manager need to implement the
    :meth:`.transform_entities` method, which will convert the supported
    entities from whichever format the integration supports to a collection of
    :class:`platypush.entities.Entity` objects.

    The converted entities will then be passed to the
    :class:`platypush.entities.EntitiesEngine` whenever
    :meth:`.publish_entities` is called

    The implemented classes should also implement the :meth:`.status` method.
    This method should retrieve the current state of the entities and call
    :meth:`.publish_entities`.
    """

    def __new__(cls, *_, **__) -> 'EntityManager':
        register_entity_manager(cls)
        return super().__new__(cls)

    @abstractmethod
    def transform_entities(self, entities: Collection[Any], **_) -> Collection[Entity]:
        """
        This method takes a list of entities in any (plugin-specific)
        format and converts them into a standardized collection of
        `Entity` objects. Since this method is called by
        :meth:`.publish_entities` before entity updates are published,
        you may usually want to extend it to pre-process the entities
        managed by your extension into the standard format before they
        are stored and published to all the consumers.
        """
        assert all(isinstance(e, Entity) for e in entities), (
            'Expected all the instances to be entities, got '
            f'{[e.__class__.__name__ for e in entities]}'
        )
        return entities

    @abstractmethod
    def status(self, *_, **__):
        """
        All derived classes should implement this method.

        At the very least, this method should refresh the current state of the
        integration's entities and call :meth:`.publish_entities`.

        It should also return the current state of the entities as a list of
        serialized entities, if possible.
        """
        raise NotImplementedError(
            'The `status` method has not been implemented in '
            f'{self.__class__.__name__}'
        )

    def _normalize_entities(self, entities: Collection[Entity]) -> Collection[Entity]:
        for entity in entities:
            if not entity:
                continue

            if entity.id and not entity.external_id:
                # Entity IDs can only refer to the internal primary key
                entity.external_id = entity.id
                entity.id = None  # type: ignore

            entity.plugin = get_plugin_name_by_class(self.__class__)  # type: ignore
            entity.updated_at = utcnow()  # type: ignore
            entity.children = self._normalize_entities(entity.children)

        return entities

    def publish_entities(
        self,
        entities: Optional[Collection[Any]],
        callback: Optional[EntitySavedCallback] = None,
        **kwargs,
    ) -> Collection[Entity]:
        """
        Publishes a list of entities. The downstream consumers include:

            - The entity persistence manager
            - The web server
            - Any consumer subscribed to
                :class:`platypush.message.event.entities.EntityUpdateEvent`
                events (e.g. web clients)

        It also accepts an optional callback that will be called when each of
        the entities in the set is flushed to the database.

        You usually don't need to override this class (but you may want to
        extend :meth:`.transform_entities` instead if your extension doesn't
        natively handle `Entity` objects).
        """
        from platypush.entities import publish_entities

        transformed_entities = self._normalize_entities(
            self.transform_entities(entities or [], **kwargs)
        )

        publish_entities(transformed_entities, callback=callback)
        return transformed_entities


def register_entity_manager(cls: Type[EntityManager]):
    """
    Associates a plugin as a manager for a certain entity type.
    You usually don't have to call this method directly.
    """
    entity_managers = [c for c in inspect.getmro(cls) if issubclass(c, EntityManager)]

    plugin_name = get_plugin_name_by_class(cls) or ''
    redis = get_redis()
    registry = get_plugin_entity_registry()
    registry_by_plugin = set(registry['by_plugin'].get(plugin_name, []))

    for manager in entity_managers:
        entity_type_name = manager.__name__
        registry_by_type = set(registry['by_type'].get(entity_type_name, []))
        registry_by_plugin.add(entity_type_name)
        registry_by_type.add(plugin_name)
        registry['by_plugin'][plugin_name] = list(registry_by_plugin)
        registry['by_type'][entity_type_name] = list(registry_by_type)

    redis.mset({_entity_registry_varname: json.dumps(registry)})


def get_plugin_entity_registry() -> Dict[str, Dict[str, Collection[str]]]:
    """
    Get the `plugin->entity_types` and `entity_type->plugin`
    mappings supported by the current configuration.
    """
    redis = get_redis()
    registry = redis.mget([_entity_registry_varname])[0]
    try:
        registry = json.loads((registry or b'').decode())
    except (TypeError, ValueError):
        return {'by_plugin': {}, 'by_type': {}}

    enabled_plugins = set(Config.get_plugins().keys())

    return {
        'by_plugin': {
            plugin_name: entity_types
            for plugin_name, entity_types in registry.get('by_plugin', {}).items()
            if plugin_name in enabled_plugins
        },
        'by_type': {
            entity_type: [p for p in plugins if p in enabled_plugins]
            for entity_type, plugins in registry.get('by_type', {}).items()
        },
    }



from typing import Dict, Iterable, Optional, Tuple

from bleak.backends.device import BLEDevice

from .._cache import BaseCache


class DeviceCache(BaseCache):
    """
    Cache used to store scanned Bluetooth devices as :class:`BLEDevice`.
    """

    _by_address: Dict[str, BLEDevice]
    _by_name: Dict[str, BLEDevice]

    @property
    def _address_field(self) -> str:
        return 'address'

    @property
    def _name_field(self) -> str:
        return 'name'

    def get(self, device: str) -> Optional[BLEDevice]:
        return super().get(device)

    def add(self, device: BLEDevice) -> BLEDevice:
        return super().add(device)

    def values(self) -> Iterable[BLEDevice]:
        return super().values()

    def items(self) -> Iterable[Tuple[str, BLEDevice]]:
        return super().items()



import importlib
import logging
import re
import textwrap as tw
from typing import Callable, Union

from platypush.utils import get_defining_class

from .._model.constants import doc_base_url


# pylint: disable=too-few-public-methods
class RstExtensionsMixin:
    """
    Mixin class for handling non-standard reStructuredText extensions.
    """

    _rst_extensions = {
        name: re.compile(regex)
        for name, regex in {
            "class": "(:class:`(?P<name>[^`]+)`)",
            "method": "(:meth:`(?P<name>[^`]+)`)",
            "module": "(:mod:`(?P<name>[^`]+)`)",
            "function": "(:func:`(?P<name>[^`]+)`)",
            "schema": r"^((?P<indent>\s*)(?P<before>.*)"
            r"(\.\. schema:: (?P<name>[\w.]+)\s*"
            r"(\((?P<args>.+?)\))?)(?P<after>.*))$",
        }.items()
    }

    logger = logging.getLogger(__name__)

    @classmethod
    def _expand_rst_extensions(cls, docstr: str, obj: Union[Callable, type]) -> str:
        """
        Expand the known reStructuredText extensions in a docstring.
        """
        for ex_name, regex in cls._rst_extensions.items():
            while True:
                match = regex.search(docstr)
                if not match:
                    break

                try:
                    docstr = cls._expand_rst_extension(docstr, ex_name, match, obj)
                except Exception as e:
                    cls.logger.warning(
                        "Could not import module %s: %s", match.group("name"), e
                    )

                    cls.logger.exception(e)

        return docstr

    @classmethod
    def _expand_rst_extension(
        cls, docstr: str, ex_name: str, match: re.Match[str], obj: Union[Callable, type]
    ) -> str:
        if ex_name == "schema":
            return cls._expand_schema(docstr, match)

        return cls._expand_module(docstr, ex_name, match, obj)

    @staticmethod
    def _field_default(field):
        return getattr(field, "dump_default", getattr(field, "default", None))

    @classmethod
    def _expand_schema(cls, docstr: str, match: re.Match) -> str:
        from marshmallow import missing
        from marshmallow.validate import OneOf

        value = match.group("name")
        mod = importlib.import_module(
            "platypush.schemas." + ".".join(value.split(".")[:-1])
        )
        obj_cls = getattr(mod, value.split(".")[-1])
        schema_args = (
            eval(f'dict({match.group("args")})') if match.group("args") else {}
        )
        obj = obj_cls(**schema_args)

        schema_doc = tw.indent(
            ".. code-block:: python\n\n"
            + tw.indent(
                ("[" if obj.many else "")
                + "{\n"
                + tw.indent(
                    ",\n".join(
                        (
                            (
                                "# " + field.metadata["description"] + "\n"
                                if field.metadata.get("description")
                                else ""
                            )
                            + (
                                "# Possible values: "
                                + str(field.validate.choices)
                                + "\n"
                                if isinstance(field.validate, OneOf)
                                else ""
                            )
                            + f'"{field_name}": '
                            + (
                                (
                                    '"'
                                    + field.metadata.get(
                                        "example", cls._field_default(field)
                                    )
                                    + '"'
                                    if isinstance(
                                        field.metadata.get(
                                            "example", cls._field_default(field)
                                        ),
                                        str,
                                    )
                                    else str(
                                        field.metadata.get(
                                            "example", cls._field_default(field)
                                        )
                                    )
                                )
                                if not (
                                    field.metadata.get("example") is None
                                    and cls._field_default(field) is missing
                                )
                                else "..."
                            )
                        )
                        for field_name, field in obj.fields.items()
                    ),
                    prefix="  ",
                )
                + "\n}"
                + ("]" if obj.many else ""),
                prefix="  ",
            ),
            prefix=match.group("indent") + "  ",
        )

        docstr = docstr.replace(
            match.group(0),
            match.group("before") + "\n\n" + schema_doc + "\n\n" + match.group("after"),
        )

        return docstr

    @classmethod
    def _expand_module(
        cls,
        docstr: str,
        ex_name: str,
        match: re.Match,
        obj: Union[Callable, type],
    ) -> str:
        value = match.group("name")
        modname = obj_name = url_path = None

        if ex_name == "module":
            modname = obj_name = value
        elif value.startswith("."):
            base_cls = get_defining_class(obj)
            if base_cls:
                modname = base_cls.__module__
                obj_name = f'{base_cls.__qualname__}.{value[1:]}'
        elif ex_name == "method":
            modname = ".".join(value.split(".")[:-2])
            obj_name = ".".join(value.split(".")[-2:])
        else:
            modname = ".".join(value.split(".")[:-1])
            obj_name = value.split(".")[-1]

        if modname and obj_name:
            if modname.startswith("platypush.plugins"):
                url_path = "plugins/" + ".".join(modname.split(".")[2:])
            elif modname.startswith("platypush.backend"):
                url_path = "backend/" + ".".join(modname.split(".")[2:])
            elif modname.startswith("platypush.message.event"):
                url_path = "events/" + ".".join(modname.split(".")[3:])
            elif modname.startswith("platypush.message.response"):
                url_path = "responses/" + ".".join(modname.split(".")[3:])

        if url_path:
            docstr = docstr.replace(
                match.group(0),
                f"`{obj_name} <{doc_base_url}/{url_path}.html"
                + (f"#{modname}.{obj_name}" if ex_name != "module" else "")
                + ">`_",
            )
        else:
            docstr = docstr.replace(match.group(0), f"``{value}``")

        return docstr



from abc import ABC, abstractmethod
from typing import Union
from uuid import UUID

from . import EntityManager

InstanceId = Union[str, int, UUID]


class CloudInstanceEntityManager(EntityManager, ABC):
    """
    Base class for integrations that support cloud instances (like Linode or
    AWS).
    """

    @abstractmethod
    def reboot(self, instance: InstanceId, **_):
        """
        Reboot an instance.

        :param instance: ID or name of the instance to be rebooted.
        """
        raise NotImplementedError()

    @abstractmethod
    def boot(self, instance: InstanceId, **_):
        """
        Boot an instance.

        :param instance: ID or name of the instance to be rebooted.
        """
        raise NotImplementedError()

    @abstractmethod
    def shutdown(self, instance: InstanceId, **_):
        """
        Shutdown an instance.

        :param instance: ID or name of the instance to be rebooted.
        """
        raise NotImplementedError()
